from .exceptions import *
from .options import *
from .module import *
from .lifecycle import *

        
class App(Module):
    def __init__(self, *args):
        self._provides: dict = {}
        self._invokes: [InvokeTarget] = []
        self.provided: dict = {}
        self.lifecycle = Lifecycle()

        for option in args:
            if not isinstance(option, Option):
                raise PyDITypeError(
                    "App constructor arguments must be of type Option")
            option.apply(self)

    def build_dependencies(self, target: ProvideTarget | InvokeTarget) -> []:
        inject = []
        for require in target.requires:
            if require == Lifecycle:
                inject.append(self.lifecycle)
                continue

            key = require
            if isinstance(require, NamedProvider):
                key = require.name

            if key not in self._provides:
                raise MissingDependencyError("Cannot find dependency", key)

            if key in self.provided:
                if isinstance(require, NamedProvider) and require.provider != self.provided[key].__class__:
                    raise DependencyTypeError("Dependency is of wrong type", key,
                                              self.provided[key].__class__, "but should be of type", require.provider)
                inject.append(self.provided[key])
            else:
                provider = self.build_provider(self._provides[key])
                if isinstance(require, NamedProvider) and require.provider != provider.__class__:
                    raise DependencyTypeError("Dependency is of wrong type", key,
                                              provider.__class__, "but should be of type", require.provider)
                self.provided[key] = provider
                inject.append(provider)

        return inject

    def build_provider(self, target: ProvideTarget) -> object:
        inject = self.build_dependencies(target)
        try:
            return target.callable(*inject)
        except TypeError as e:
                if "missing" in str(e):
                    raise MissingHintError(str(e), "verify if all hints are set")
                raise e

    def run(self):
        for invoke in self._invokes:
            inject = self.build_dependencies(invoke)
            try:
                invoke.callable(*inject)
            except TypeError as e:
                if "missing" in str(e):
                    raise MissingHintError(str(e), "verify if all hints are set")
                raise e
            self.lifecycle.start()



import typing


from .app import *
from .module import *



class NamedProvider:
    def __init__(self, name: str, provider: object):
        self.name = name
        self.provider = provider



class Option:
    def apply(self, mod: Module):
        pass


class ProvideOption(Option):
    def __init__(self, *args):
        self._targets = {}
        for arg in args:
            if isinstance(arg, NamedProvider):
                hints = typing.get_type_hints(arg.provider.__init__)
                self._targets[arg.name] = ProvideTarget(
                    callable=arg.provider,
                    provides=arg.name,
                    requires=get_requires_from_hints(hints))
                continue

            # check if it is a class of any type
            if isinstance(arg, type):
                hints = typing.get_type_hints(arg.__init__)
                self._targets[arg] = ProvideTarget(
                    callable=arg,
                    provides=arg,
                    requires=get_requires_from_hints(hints))
                continue

            if callable(arg):
                hints = typing.get_type_hints(arg)
                if "return" not in hints:
                    raise MissingHintError(
                        "Provide target must have a return type hint")
                self._targets[hints["return"]] = ProvideTarget(
                    callable=arg,
                    provides=hints["return"],
                    requires=get_requires_from_hints(hints))
                continue

            raise PyDITypeError(
                "Provide target must be a callable constructor, a class or a NamedProvider")

    def apply(self, mod: Module):
        mod.add_provides(self._targets)

def get_requires_from_hints(hints) -> [str]:
    requires = []
    for key, value in hints.items():
        if key == "return":
            continue
        else:
            requires.append(value)
    return requires


class InvokeOption(Option):
    def __init__(self, *args):
        self._targets = []
        for arg in args:
            if not callable(arg):
                raise PyDITypeError("Provide target must be callable")

            hints = typing.get_type_hints(arg)

            self._targets.append(InvokeTarget(
                callable=arg,
                requires=get_requires_from_hints(hints)))

    def apply(self, mod: Module):
        mod.add_invokes(self._targets)



from .exceptions import *


class ProvideTarget:
    def __init__(self, callable: callable, provides: str, requires: [str]):
        self.callable = callable
        self.provides = provides
        self.requires = requires


class InvokeTarget:
    def __init__(self, callable: callable, requires: [str]):
        self.callable = callable
        self.requires = requires


class Module:
    def __init__(self) -> None:
        self._provides = {}
        self._invokes = []

    def add_provides(self, targets: dict):
        if not isinstance(targets, dict):
            raise PyDITypeError("Provide targets must be a dict")
        for key, value in targets.items():
            if not isinstance(value, ProvideTarget):
                raise PyDITypeError(
                    "Provide target must be of type ProvideTarget")
        self._provides = self._provides | targets

    def add_invokes(self, targets: [InvokeTarget]):
        for target in targets:
            if not isinstance(target, InvokeTarget):
                raise PyDITypeError(
                    "Invoke target must be of type InvokeTarget")
        self._invokes.extend(targets)



import unittest

from src.dinjections.app import *


class TestClass0:
    pass


class TestClass1:
    pass


class TestClass2:
    pass


class TestClass3:
    pass

    def run(self):
        pass


def new_test_class_3(p: NamedProvider("t1", TestClass1), p2: NamedProvider("t2", TestClass1), r: TestClass2) -> TestClass3:
    return TestClass3()


def new_test_class_3_no_hint(p: NamedProvider("t1", TestClass1), p2: NamedProvider("t2", TestClass1), r: TestClass2):
    return TestClass3()


def new_test_class_3_no_hint_input(p: NamedProvider("t1", TestClass1), p2: NamedProvider("t2", TestClass1), r) -> TestClass3:
    return TestClass3()


def invoker(l: Lifecycle, t: TestClass3):
    l.append_hook(Hook(
        on_start=lambda: {
            t.run()
        },
    ))

def invoker_no_hint(l: Lifecycle, t):
    l.append_hook(Hook(
        on_start=lambda: {
            t.run()
        },
    ))

class TestApp(unittest.TestCase):
    
    def test_app(self):
        try:
            app = App(
                ProvideOption(
                    # test named provider
                    NamedProvider("t1", TestClass1),
                    NamedProvider("t2", TestClass1),
                ),
                # test modular implementation
                ProvideOption(
                    TestClass2,
                    new_test_class_3,
                ),
                InvokeOption(
                    invoker,
                ))
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_missing_dependency(self):
        exc = None
        try:
            provides = ProvideOption()
            app = App(
                provides,
                InvokeOption(
                    invoker,
                ))
        
            app.run()
        except Exception as e:
            exc=e
        
        self.assertIsInstance(exc, MissingDependencyError)

    def test_dependency_type_error(self):
        exc = None
        try:
            provides = ProvideOption(
                    NamedProvider("t1", TestClass1),
                    NamedProvider("t2", TestClass2),
                    TestClass2,
                    new_test_class_3,
            )
            app = App(
                provides,
                InvokeOption(
                    invoker,
                ))
        
            app.run()
        except Exception as e:
            exc=e
        
        self.assertIsInstance(exc, DependencyTypeError)

    def test_missing_hint_error(self):
        exc = None
        try:
            provides = ProvideOption(
                    NamedProvider("t1", TestClass1),
                    NamedProvider("t2", TestClass1),
                    TestClass2,
                    new_test_class_3_no_hint,
            )
            app = App(
                provides,
                InvokeOption(
                    invoker,
                ))
        
            app.run()
        except Exception as e:
            exc=e
        
        self.assertIsInstance(exc, MissingHintError)

    def test_missing_hint_error_input_invoke(self):
        exc = None
        try:
            provides = ProvideOption(
                    NamedProvider("t1", TestClass1),
                    NamedProvider("t2", TestClass1),
                    TestClass2,
                    new_test_class_3,
            )
            app = App(
                provides,
                InvokeOption(
                    invoker_no_hint,
                ))
        
            app.run()
        except Exception as e:
            exc=e
        
        self.assertIsInstance(exc, MissingHintError)

    def test_missing_hint_error_input_provide(self):
        exc = None
        try:
            provides = ProvideOption(
                    NamedProvider("t1", TestClass1),
                    NamedProvider("t2", TestClass1),
                    TestClass2,
                    new_test_class_3_no_hint_input,
            )
            app = App(
                provides,
                InvokeOption(
                    invoker,
                ))
        
            app.run()
        except Exception as e:
            exc=e
        
        self.assertIsInstance(exc, MissingHintError)


if __name__ == '__main__':
    unittest.main()
